from opentelemetry import trace
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor
from opentelemetry.sdk.resources import SERVICE_NAME, Resource
from opentelemetry.exporter.otlp.proto.http.trace_exporter import OTLPSpanExporter


class Tracer:
    _instance = None

    def __init__(
        self,
        service_name="maihem-workflow",
        endpoint="https://ingest.maihem.com/v1/traces",
    ):
        resource_attrs = {SERVICE_NAME: service_name}

        resource = Resource(attributes=resource_attrs)
        self.provider = TracerProvider(resource=resource)

        # Use OTLP HTTP exporter with BatchSpanProcessor
        otlp_exporter = OTLPSpanExporter(endpoint=endpoint)
        self.processor = BatchSpanProcessor(
            otlp_exporter,
            schedule_delay_millis=5 * 60 * 1000,
            export_timeout_millis=5 * 60 * 1000 + 1000,
        )
        self.provider.add_span_processor(self.processor)

        trace.set_tracer_provider(self.provider)
        self.tracer = trace.get_tracer(service_name)

    @classmethod
    def get_instance(cls):
        if cls._instance is None:
            cls._instance = Tracer()
        return cls._instance


def set_attribute(key: str, value: str):
    """Set an attribute on the current Maihem span."""
    current_span = trace.get_current_span()
    current_span.set_attribute(key, value)



import datetime
from typing import Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

T = TypeVar("T", bound="Dataset")


@_attrs_define
class Dataset:
    """
    Attributes:
        id (str):
        created_at (datetime.datetime):
        updated_at (datetime.datetime):
        org_id (str):
        name (str):
        label (Union[None, str]):
        description (Union[None, str]):
        source_type (str):
        source_reference (Union[None, str]):
        external_id (Union[None, str]):
        is_deleted (bool):
        target_type (str):
        target_id (Union[None, str]):
    """

    id: str
    created_at: datetime.datetime
    updated_at: datetime.datetime
    org_id: str
    name: str
    label: Union[None, str]
    description: Union[None, str]
    source_type: str
    source_reference: Union[None, str]
    external_id: Union[None, str]
    is_deleted: bool
    target_type: str
    target_id: Union[None, str]
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        id = self.id

        created_at = self.created_at.isoformat()

        updated_at = self.updated_at.isoformat()

        org_id = self.org_id

        name = self.name

        label: Union[None, str]
        label = self.label

        description: Union[None, str]
        description = self.description

        source_type = self.source_type

        source_reference: Union[None, str]
        source_reference = self.source_reference

        external_id: Union[None, str]
        external_id = self.external_id

        is_deleted = self.is_deleted

        target_type = self.target_type

        target_id: Union[None, str]
        target_id = self.target_id

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "id": id,
                "created_at": created_at,
                "updated_at": updated_at,
                "org_id": org_id,
                "name": name,
                "label": label,
                "description": description,
                "source_type": source_type,
                "source_reference": source_reference,
                "external_id": external_id,
                "is_deleted": is_deleted,
                "target_type": target_type,
                "target_id": target_id,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        d = src_dict.copy()
        id = d.pop("id")

        created_at = isoparse(d.pop("created_at"))

        updated_at = isoparse(d.pop("updated_at"))

        org_id = d.pop("org_id")

        name = d.pop("name")

        def _parse_label(data: object) -> Union[None, str]:
            if data is None:
                return data
            return cast(Union[None, str], data)

        label = _parse_label(d.pop("label"))

        def _parse_description(data: object) -> Union[None, str]:
            if data is None:
                return data
            return cast(Union[None, str], data)

        description = _parse_description(d.pop("description"))

        source_type = d.pop("source_type")

        def _parse_source_reference(data: object) -> Union[None, str]:
            if data is None:
                return data
            return cast(Union[None, str], data)

        source_reference = _parse_source_reference(d.pop("source_reference"))

        def _parse_external_id(data: object) -> Union[None, str]:
            if data is None:
                return data
            return cast(Union[None, str], data)

        external_id = _parse_external_id(d.pop("external_id"))

        is_deleted = d.pop("is_deleted")

        target_type = d.pop("target_type")

        def _parse_target_id(data: object) -> Union[None, str]:
            if data is None:
                return data
            return cast(Union[None, str], data)

        target_id = _parse_target_id(d.pop("target_id"))

        dataset = cls(
            id=id,
            created_at=created_at,
            updated_at=updated_at,
            org_id=org_id,
            name=name,
            label=label,
            description=description,
            source_type=source_type,
            source_reference=source_reference,
            external_id=external_id,
            is_deleted=is_deleted,
            target_type=target_type,
            target_id=target_id,
        )

        dataset.additional_properties = d
        return dataset

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import datetime
from typing import Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..models.evaluation_feedback_feedback import EvaluationFeedbackFeedback
from ..types import UNSET, Unset

T = TypeVar("T", bound="EvaluationFeedback")


@_attrs_define
class EvaluationFeedback:
    """
    Attributes:
        id (str):
        created_at (datetime.datetime):
        updated_at (datetime.datetime):
        entity_id (str):
        entity_type (str):
        created_by (str):
        feedback (EvaluationFeedbackFeedback):
        evaluation_id (Union[None, Unset, str]):
        comments (Union[None, Unset, str]):
    """

    id: str
    created_at: datetime.datetime
    updated_at: datetime.datetime
    entity_id: str
    entity_type: str
    created_by: str
    feedback: EvaluationFeedbackFeedback
    evaluation_id: Union[None, Unset, str] = UNSET
    comments: Union[None, Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        id = self.id

        created_at = self.created_at.isoformat()

        updated_at = self.updated_at.isoformat()

        entity_id = self.entity_id

        entity_type = self.entity_type

        created_by = self.created_by

        feedback = self.feedback.value

        evaluation_id: Union[None, Unset, str]
        if isinstance(self.evaluation_id, Unset):
            evaluation_id = UNSET
        else:
            evaluation_id = self.evaluation_id

        comments: Union[None, Unset, str]
        if isinstance(self.comments, Unset):
            comments = UNSET
        else:
            comments = self.comments

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "id": id,
                "created_at": created_at,
                "updated_at": updated_at,
                "entity_id": entity_id,
                "entity_type": entity_type,
                "created_by": created_by,
                "feedback": feedback,
            }
        )
        if evaluation_id is not UNSET:
            field_dict["evaluation_id"] = evaluation_id
        if comments is not UNSET:
            field_dict["comments"] = comments

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        d = src_dict.copy()
        id = d.pop("id")

        created_at = isoparse(d.pop("created_at"))

        updated_at = isoparse(d.pop("updated_at"))

        entity_id = d.pop("entity_id")

        entity_type = d.pop("entity_type")

        created_by = d.pop("created_by")

        feedback = EvaluationFeedbackFeedback(d.pop("feedback"))

        def _parse_evaluation_id(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        evaluation_id = _parse_evaluation_id(d.pop("evaluation_id", UNSET))

        def _parse_comments(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        comments = _parse_comments(d.pop("comments", UNSET))

        evaluation_feedback = cls(
            id=id,
            created_at=created_at,
            updated_at=updated_at,
            entity_id=entity_id,
            entity_type=entity_type,
            created_by=created_by,
            feedback=feedback,
            evaluation_id=evaluation_id,
            comments=comments,
        )

        evaluation_feedback.additional_properties = d
        return evaluation_feedback

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import sys, os
from loguru import logger
from typing import TYPE_CHECKING
from .request_context import (
    get_request_id,
    get_test_run_id,
    get_conversation_id,
    get_workflow_span_id,
    get_workflow_trace_id,
    get_workflow_trace_raw_id,
    get_workflow_step_id,
)

if TYPE_CHECKING:
    from loguru import Logger

from loguru._logger import Logger

extra_fields = [
    "request_id",
    "test_run_id",
    "conversation_id",
    "workflow_trace_id",
    "workflow_trace_raw_id",
    "workflow_span_id",
    "workflow_step_id",
]


def _add_extra_fields(record):
    record["extra"]["request_id"] = get_request_id()
    record["extra"]["test_run_id"] = get_test_run_id()
    record["extra"]["conversation_id"] = get_conversation_id()
    record["extra"]["workflow_trace_id"] = get_workflow_trace_id()
    record["extra"]["workflow_trace_raw_id"] = get_workflow_trace_raw_id()
    record["extra"]["workflow_span_id"] = get_workflow_span_id()
    record["extra"]["workflow_step_id"] = get_workflow_step_id()
    return True


def add_default_logger(logger: Logger, level: str = "DEBUG"):
    logger.remove()

    def format_message(record):
        # Add a relative pathname to 'extra', but do not overwrite existing keys
        record["extra"][
            "pathname"
        ] = f"{os.path.relpath(record['file'].path)}:{record['line']}"

        # Base format string
        base_format = (
            "<blue>{time:YYYY-MM-DD HH:mm:ss}</blue> - "
            "<level>{level}</level> - {message} - "
            "<dim>{extra[pathname]}</dim>"
        )

        # Add any additional 'extra' fields dynamically
        for key, value in record["extra"].items():
            if key == "pathname":  # Skip 'pathname' since it's already included
                continue
            if value:  # Only include non-empty values
                base_format += f" - <dim>{key}:{{extra[{key}]}}</dim>"

        # Include exception details if present
        if record["exception"]:
            base_format += "\n<red>{exception}</red>"

        base_format += "\n"
        return base_format

    logger.add(
        sys.stdout,
        colorize=True,
        format=format_message,
        level=level,
        filter=_add_extra_fields,
        backtrace=True,
        diagnose=False,
    )


def get_logger() -> Logger:
    return logger



import datetime
from typing import TYPE_CHECKING, Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..models.agent_type import AgentType
from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.conversation_nested_evaluation import ConversationNestedEvaluation
    from ..models.conversation_nested_sentence import ConversationNestedSentence
    from ..models.conversation_nested_workflow_trace import ConversationNestedWorkflowTrace


T = TypeVar("T", bound="ConversationNestedMessage")


@_attrs_define
class ConversationNestedMessage:
    """
    Attributes:
        id (str):
        created_at (datetime.datetime):
        updated_at (datetime.datetime):
        agent_type (AgentType):
        content (Union[None, Unset, str]):
        end_code (Union[None, Unset, str]):
        evaluations (Union[Unset, list['ConversationNestedEvaluation']]):
        conversation_sentences (Union[Unset, list['ConversationNestedSentence']]):
        workflow_trace (Union['ConversationNestedWorkflowTrace', None, Unset]):
    """

    id: str
    created_at: datetime.datetime
    updated_at: datetime.datetime
    agent_type: AgentType
    content: Union[None, Unset, str] = UNSET
    end_code: Union[None, Unset, str] = UNSET
    evaluations: Union[Unset, list["ConversationNestedEvaluation"]] = UNSET
    conversation_sentences: Union[Unset, list["ConversationNestedSentence"]] = UNSET
    workflow_trace: Union["ConversationNestedWorkflowTrace", None, Unset] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        from ..models.conversation_nested_workflow_trace import ConversationNestedWorkflowTrace

        id = self.id

        created_at = self.created_at.isoformat()

        updated_at = self.updated_at.isoformat()

        agent_type = self.agent_type.value

        content: Union[None, Unset, str]
        if isinstance(self.content, Unset):
            content = UNSET
        else:
            content = self.content

        end_code: Union[None, Unset, str]
        if isinstance(self.end_code, Unset):
            end_code = UNSET
        else:
            end_code = self.end_code

        evaluations: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.evaluations, Unset):
            evaluations = []
            for evaluations_item_data in self.evaluations:
                evaluations_item = evaluations_item_data.to_dict()
                evaluations.append(evaluations_item)

        conversation_sentences: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.conversation_sentences, Unset):
            conversation_sentences = []
            for conversation_sentences_item_data in self.conversation_sentences:
                conversation_sentences_item = conversation_sentences_item_data.to_dict()
                conversation_sentences.append(conversation_sentences_item)

        workflow_trace: Union[None, Unset, dict[str, Any]]
        if isinstance(self.workflow_trace, Unset):
            workflow_trace = UNSET
        elif isinstance(self.workflow_trace, ConversationNestedWorkflowTrace):
            workflow_trace = self.workflow_trace.to_dict()
        else:
            workflow_trace = self.workflow_trace

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "id": id,
                "created_at": created_at,
                "updated_at": updated_at,
                "agent_type": agent_type,
            }
        )
        if content is not UNSET:
            field_dict["content"] = content
        if end_code is not UNSET:
            field_dict["end_code"] = end_code
        if evaluations is not UNSET:
            field_dict["evaluations"] = evaluations
        if conversation_sentences is not UNSET:
            field_dict["conversation_sentences"] = conversation_sentences
        if workflow_trace is not UNSET:
            field_dict["workflow_trace"] = workflow_trace

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        from ..models.conversation_nested_evaluation import ConversationNestedEvaluation
        from ..models.conversation_nested_sentence import ConversationNestedSentence
        from ..models.conversation_nested_workflow_trace import ConversationNestedWorkflowTrace

        d = src_dict.copy()
        id = d.pop("id")

        created_at = isoparse(d.pop("created_at"))

        updated_at = isoparse(d.pop("updated_at"))

        agent_type = AgentType(d.pop("agent_type"))

        def _parse_content(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        content = _parse_content(d.pop("content", UNSET))

        def _parse_end_code(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        end_code = _parse_end_code(d.pop("end_code", UNSET))

        evaluations = []
        _evaluations = d.pop("evaluations", UNSET)
        for evaluations_item_data in _evaluations or []:
            evaluations_item = ConversationNestedEvaluation.from_dict(evaluations_item_data)

            evaluations.append(evaluations_item)

        conversation_sentences = []
        _conversation_sentences = d.pop("conversation_sentences", UNSET)
        for conversation_sentences_item_data in _conversation_sentences or []:
            conversation_sentences_item = ConversationNestedSentence.from_dict(conversation_sentences_item_data)

            conversation_sentences.append(conversation_sentences_item)

        def _parse_workflow_trace(data: object) -> Union["ConversationNestedWorkflowTrace", None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                workflow_trace_type_0 = ConversationNestedWorkflowTrace.from_dict(data)

                return workflow_trace_type_0
            except:  # noqa: E722
                pass
            return cast(Union["ConversationNestedWorkflowTrace", None, Unset], data)

        workflow_trace = _parse_workflow_trace(d.pop("workflow_trace", UNSET))

        conversation_nested_message = cls(
            id=id,
            created_at=created_at,
            updated_at=updated_at,
            agent_type=agent_type,
            content=content,
            end_code=end_code,
            evaluations=evaluations,
            conversation_sentences=conversation_sentences,
            workflow_trace=workflow_trace,
        )

        conversation_nested_message.additional_properties = d
        return conversation_nested_message

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Literal, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.dataset_create_request_target_type import DatasetCreateRequestTargetType
from ..types import UNSET, Unset

T = TypeVar("T", bound="DatasetCreateRequest")


@_attrs_define
class DatasetCreateRequest:
    """
    Attributes:
        name (str):
        label (Union[None, Unset, str]):
        description (Union[None, Unset, str]):
        source_type (Union[Literal['file'], Unset]):  Default: 'file'.
        source_reference (Union[None, Unset, str]):
        target_type (Union[Unset, DatasetCreateRequestTargetType]):  Default: DatasetCreateRequestTargetType.WORKFLOW.
        target_id (Union[None, Unset, str]):
        external_id (Union[None, Unset, str]):
    """

    name: str
    label: Union[None, Unset, str] = UNSET
    description: Union[None, Unset, str] = UNSET
    source_type: Union[Literal["file"], Unset] = "file"
    source_reference: Union[None, Unset, str] = UNSET
    target_type: Union[Unset, DatasetCreateRequestTargetType] = DatasetCreateRequestTargetType.WORKFLOW
    target_id: Union[None, Unset, str] = UNSET
    external_id: Union[None, Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        name = self.name

        label: Union[None, Unset, str]
        if isinstance(self.label, Unset):
            label = UNSET
        else:
            label = self.label

        description: Union[None, Unset, str]
        if isinstance(self.description, Unset):
            description = UNSET
        else:
            description = self.description

        source_type = self.source_type

        source_reference: Union[None, Unset, str]
        if isinstance(self.source_reference, Unset):
            source_reference = UNSET
        else:
            source_reference = self.source_reference

        target_type: Union[Unset, str] = UNSET
        if not isinstance(self.target_type, Unset):
            target_type = self.target_type.value

        target_id: Union[None, Unset, str]
        if isinstance(self.target_id, Unset):
            target_id = UNSET
        else:
            target_id = self.target_id

        external_id: Union[None, Unset, str]
        if isinstance(self.external_id, Unset):
            external_id = UNSET
        else:
            external_id = self.external_id

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "name": name,
            }
        )
        if label is not UNSET:
            field_dict["label"] = label
        if description is not UNSET:
            field_dict["description"] = description
        if source_type is not UNSET:
            field_dict["source_type"] = source_type
        if source_reference is not UNSET:
            field_dict["source_reference"] = source_reference
        if target_type is not UNSET:
            field_dict["target_type"] = target_type
        if target_id is not UNSET:
            field_dict["target_id"] = target_id
        if external_id is not UNSET:
            field_dict["external_id"] = external_id

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        d = src_dict.copy()
        name = d.pop("name")

        def _parse_label(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        label = _parse_label(d.pop("label", UNSET))

        def _parse_description(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        description = _parse_description(d.pop("description", UNSET))

        source_type = cast(Union[Literal["file"], Unset], d.pop("source_type", UNSET))
        if source_type != "file" and not isinstance(source_type, Unset):
            raise ValueError(f"source_type must match const 'file', got '{source_type}'")

        def _parse_source_reference(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        source_reference = _parse_source_reference(d.pop("source_reference", UNSET))

        _target_type = d.pop("target_type", UNSET)
        target_type: Union[Unset, DatasetCreateRequestTargetType]
        if isinstance(_target_type, Unset):
            target_type = UNSET
        else:
            target_type = DatasetCreateRequestTargetType(_target_type)

        def _parse_target_id(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        target_id = _parse_target_id(d.pop("target_id", UNSET))

        def _parse_external_id(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        external_id = _parse_external_id(d.pop("external_id", UNSET))

        dataset_create_request = cls(
            name=name,
            label=label,
            description=description,
            source_type=source_type,
            source_reference=source_reference,
            target_type=target_type,
            target_id=target_id,
            external_id=external_id,
        )

        dataset_create_request.additional_properties = d
        return dataset_create_request

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="CriteriaInstanceBase")


@_attrs_define
class CriteriaInstanceBase:
    """
    Attributes:
        instances (int):
        conversations (int):
        score_impact (Union[None, Unset, float]):
    """

    instances: int
    conversations: int
    score_impact: Union[None, Unset, float] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        instances = self.instances

        conversations = self.conversations

        score_impact: Union[None, Unset, float]
        if isinstance(self.score_impact, Unset):
            score_impact = UNSET
        else:
            score_impact = self.score_impact

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "instances": instances,
                "conversations": conversations,
            }
        )
        if score_impact is not UNSET:
            field_dict["score_impact"] = score_impact

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        d = src_dict.copy()
        instances = d.pop("instances")

        conversations = d.pop("conversations")

        def _parse_score_impact(data: object) -> Union[None, Unset, float]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, float], data)

        score_impact = _parse_score_impact(d.pop("score_impact", UNSET))

        criteria_instance_base = cls(
            instances=instances,
            conversations=conversations,
            score_impact=score_impact,
        )

        criteria_instance_base.additional_properties = d
        return criteria_instance_base

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import TYPE_CHECKING, Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.agent_type import AgentType
from ..models.test_create_request_entity_type import TestCreateRequestEntityType
from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.test_create_request_documents_type_0 import TestCreateRequestDocumentsType0
    from ..models.test_create_request_metrics_config_type_0 import TestCreateRequestMetricsConfigType0


T = TypeVar("T", bound="TestCreateRequest")


@_attrs_define
class TestCreateRequest:
    """
    Attributes:
        name (str):
        agent_target_id (str):
        entity_id (str):
        label (Union[None, Unset, str]):
        initiating_agent (Union[Unset, AgentType]):
        conversation_turns_max (Union[None, Unset, int]):
        agent_maihem_behavior_prompt (Union[None, Unset, str]):
        agent_maihem_goal_prompt (Union[None, Unset, str]):
        agent_maihem_population_prompt (Union[None, Unset, str]):
        metrics_config (Union['TestCreateRequestMetricsConfigType0', None, Unset]):
        dataset_id (Union[None, Unset, str]):
        documents (Union['TestCreateRequestDocumentsType0', None, Unset]):
        is_dev_mode (Union[None, Unset, bool]):  Default: False.
        entity_type (Union[Unset, TestCreateRequestEntityType]):  Default: TestCreateRequestEntityType.WORKFLOW.
    """

    name: str
    agent_target_id: str
    entity_id: str
    label: Union[None, Unset, str] = UNSET
    initiating_agent: Union[Unset, AgentType] = UNSET
    conversation_turns_max: Union[None, Unset, int] = UNSET
    agent_maihem_behavior_prompt: Union[None, Unset, str] = UNSET
    agent_maihem_goal_prompt: Union[None, Unset, str] = UNSET
    agent_maihem_population_prompt: Union[None, Unset, str] = UNSET
    metrics_config: Union["TestCreateRequestMetricsConfigType0", None, Unset] = UNSET
    dataset_id: Union[None, Unset, str] = UNSET
    documents: Union["TestCreateRequestDocumentsType0", None, Unset] = UNSET
    is_dev_mode: Union[None, Unset, bool] = False
    entity_type: Union[Unset, TestCreateRequestEntityType] = TestCreateRequestEntityType.WORKFLOW
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        from ..models.test_create_request_documents_type_0 import TestCreateRequestDocumentsType0
        from ..models.test_create_request_metrics_config_type_0 import TestCreateRequestMetricsConfigType0

        name = self.name

        agent_target_id = self.agent_target_id

        entity_id = self.entity_id

        label: Union[None, Unset, str]
        if isinstance(self.label, Unset):
            label = UNSET
        else:
            label = self.label

        initiating_agent: Union[Unset, str] = UNSET
        if not isinstance(self.initiating_agent, Unset):
            initiating_agent = self.initiating_agent.value

        conversation_turns_max: Union[None, Unset, int]
        if isinstance(self.conversation_turns_max, Unset):
            conversation_turns_max = UNSET
        else:
            conversation_turns_max = self.conversation_turns_max

        agent_maihem_behavior_prompt: Union[None, Unset, str]
        if isinstance(self.agent_maihem_behavior_prompt, Unset):
            agent_maihem_behavior_prompt = UNSET
        else:
            agent_maihem_behavior_prompt = self.agent_maihem_behavior_prompt

        agent_maihem_goal_prompt: Union[None, Unset, str]
        if isinstance(self.agent_maihem_goal_prompt, Unset):
            agent_maihem_goal_prompt = UNSET
        else:
            agent_maihem_goal_prompt = self.agent_maihem_goal_prompt

        agent_maihem_population_prompt: Union[None, Unset, str]
        if isinstance(self.agent_maihem_population_prompt, Unset):
            agent_maihem_population_prompt = UNSET
        else:
            agent_maihem_population_prompt = self.agent_maihem_population_prompt

        metrics_config: Union[None, Unset, dict[str, Any]]
        if isinstance(self.metrics_config, Unset):
            metrics_config = UNSET
        elif isinstance(self.metrics_config, TestCreateRequestMetricsConfigType0):
            metrics_config = self.metrics_config.to_dict()
        else:
            metrics_config = self.metrics_config

        dataset_id: Union[None, Unset, str]
        if isinstance(self.dataset_id, Unset):
            dataset_id = UNSET
        else:
            dataset_id = self.dataset_id

        documents: Union[None, Unset, dict[str, Any]]
        if isinstance(self.documents, Unset):
            documents = UNSET
        elif isinstance(self.documents, TestCreateRequestDocumentsType0):
            documents = self.documents.to_dict()
        else:
            documents = self.documents

        is_dev_mode: Union[None, Unset, bool]
        if isinstance(self.is_dev_mode, Unset):
            is_dev_mode = UNSET
        else:
            is_dev_mode = self.is_dev_mode

        entity_type: Union[Unset, str] = UNSET
        if not isinstance(self.entity_type, Unset):
            entity_type = self.entity_type.value

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "name": name,
                "agent_target_id": agent_target_id,
                "entity_id": entity_id,
            }
        )
        if label is not UNSET:
            field_dict["label"] = label
        if initiating_agent is not UNSET:
            field_dict["initiating_agent"] = initiating_agent
        if conversation_turns_max is not UNSET:
            field_dict["conversation_turns_max"] = conversation_turns_max
        if agent_maihem_behavior_prompt is not UNSET:
            field_dict["agent_maihem_behavior_prompt"] = agent_maihem_behavior_prompt
        if agent_maihem_goal_prompt is not UNSET:
            field_dict["agent_maihem_goal_prompt"] = agent_maihem_goal_prompt
        if agent_maihem_population_prompt is not UNSET:
            field_dict["agent_maihem_population_prompt"] = agent_maihem_population_prompt
        if metrics_config is not UNSET:
            field_dict["metrics_config"] = metrics_config
        if dataset_id is not UNSET:
            field_dict["dataset_id"] = dataset_id
        if documents is not UNSET:
            field_dict["documents"] = documents
        if is_dev_mode is not UNSET:
            field_dict["is_dev_mode"] = is_dev_mode
        if entity_type is not UNSET:
            field_dict["entity_type"] = entity_type

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        from ..models.test_create_request_documents_type_0 import TestCreateRequestDocumentsType0
        from ..models.test_create_request_metrics_config_type_0 import TestCreateRequestMetricsConfigType0

        d = src_dict.copy()
        name = d.pop("name")

        agent_target_id = d.pop("agent_target_id")

        entity_id = d.pop("entity_id")

        def _parse_label(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        label = _parse_label(d.pop("label", UNSET))

        _initiating_agent = d.pop("initiating_agent", UNSET)
        initiating_agent: Union[Unset, AgentType]
        if isinstance(_initiating_agent, Unset):
            initiating_agent = UNSET
        else:
            initiating_agent = AgentType(_initiating_agent)

        def _parse_conversation_turns_max(data: object) -> Union[None, Unset, int]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, int], data)

        conversation_turns_max = _parse_conversation_turns_max(d.pop("conversation_turns_max", UNSET))

        def _parse_agent_maihem_behavior_prompt(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        agent_maihem_behavior_prompt = _parse_agent_maihem_behavior_prompt(d.pop("agent_maihem_behavior_prompt", UNSET))

        def _parse_agent_maihem_goal_prompt(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        agent_maihem_goal_prompt = _parse_agent_maihem_goal_prompt(d.pop("agent_maihem_goal_prompt", UNSET))

        def _parse_agent_maihem_population_prompt(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        agent_maihem_population_prompt = _parse_agent_maihem_population_prompt(
            d.pop("agent_maihem_population_prompt", UNSET)
        )

        def _parse_metrics_config(data: object) -> Union["TestCreateRequestMetricsConfigType0", None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                metrics_config_type_0 = TestCreateRequestMetricsConfigType0.from_dict(data)

                return metrics_config_type_0
            except:  # noqa: E722
                pass
            return cast(Union["TestCreateRequestMetricsConfigType0", None, Unset], data)

        metrics_config = _parse_metrics_config(d.pop("metrics_config", UNSET))

        def _parse_dataset_id(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        dataset_id = _parse_dataset_id(d.pop("dataset_id", UNSET))

        def _parse_documents(data: object) -> Union["TestCreateRequestDocumentsType0", None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                documents_type_0 = TestCreateRequestDocumentsType0.from_dict(data)

                return documents_type_0
            except:  # noqa: E722
                pass
            return cast(Union["TestCreateRequestDocumentsType0", None, Unset], data)

        documents = _parse_documents(d.pop("documents", UNSET))

        def _parse_is_dev_mode(data: object) -> Union[None, Unset, bool]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, bool], data)

        is_dev_mode = _parse_is_dev_mode(d.pop("is_dev_mode", UNSET))

        _entity_type = d.pop("entity_type", UNSET)
        entity_type: Union[Unset, TestCreateRequestEntityType]
        if isinstance(_entity_type, Unset):
            entity_type = UNSET
        else:
            entity_type = TestCreateRequestEntityType(_entity_type)

        test_create_request = cls(
            name=name,
            agent_target_id=agent_target_id,
            entity_id=entity_id,
            label=label,
            initiating_agent=initiating_agent,
            conversation_turns_max=conversation_turns_max,
            agent_maihem_behavior_prompt=agent_maihem_behavior_prompt,
            agent_maihem_goal_prompt=agent_maihem_goal_prompt,
            agent_maihem_population_prompt=agent_maihem_population_prompt,
            metrics_config=metrics_config,
            dataset_id=dataset_id,
            documents=documents,
            is_dev_mode=is_dev_mode,
            entity_type=entity_type,
        )

        test_create_request.additional_properties = d
        return test_create_request

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import datetime
from typing import TYPE_CHECKING, Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.conversation_nested_evaluation_base import ConversationNestedEvaluationBase
    from ..models.conversation_nested_workflow_span_input_payload import ConversationNestedWorkflowSpanInputPayload
    from ..models.conversation_nested_workflow_span_output_payload import ConversationNestedWorkflowSpanOutputPayload


T = TypeVar("T", bound="ConversationNestedWorkflowSpan")


@_attrs_define
class ConversationNestedWorkflowSpan:
    """
    Attributes:
        id (str):
        name (str):
        input_payload (ConversationNestedWorkflowSpanInputPayload):
        output_payload (ConversationNestedWorkflowSpanOutputPayload):
        started_at (datetime.datetime):
        completed_at (datetime.datetime):
        evaluators (Union[Unset, list[str]]):
        parent_span_id (Union[None, Unset, str]):
        evaluations (Union[Unset, list['ConversationNestedEvaluationBase']]):
    """

    id: str
    name: str
    input_payload: "ConversationNestedWorkflowSpanInputPayload"
    output_payload: "ConversationNestedWorkflowSpanOutputPayload"
    started_at: datetime.datetime
    completed_at: datetime.datetime
    evaluators: Union[Unset, list[str]] = UNSET
    parent_span_id: Union[None, Unset, str] = UNSET
    evaluations: Union[Unset, list["ConversationNestedEvaluationBase"]] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        id = self.id

        name = self.name

        input_payload = self.input_payload.to_dict()

        output_payload = self.output_payload.to_dict()

        started_at = self.started_at.isoformat()

        completed_at = self.completed_at.isoformat()

        evaluators: Union[Unset, list[str]] = UNSET
        if not isinstance(self.evaluators, Unset):
            evaluators = self.evaluators

        parent_span_id: Union[None, Unset, str]
        if isinstance(self.parent_span_id, Unset):
            parent_span_id = UNSET
        else:
            parent_span_id = self.parent_span_id

        evaluations: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.evaluations, Unset):
            evaluations = []
            for evaluations_item_data in self.evaluations:
                evaluations_item = evaluations_item_data.to_dict()
                evaluations.append(evaluations_item)

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "id": id,
                "name": name,
                "input_payload": input_payload,
                "output_payload": output_payload,
                "started_at": started_at,
                "completed_at": completed_at,
            }
        )
        if evaluators is not UNSET:
            field_dict["evaluators"] = evaluators
        if parent_span_id is not UNSET:
            field_dict["parent_span_id"] = parent_span_id
        if evaluations is not UNSET:
            field_dict["evaluations"] = evaluations

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        from ..models.conversation_nested_evaluation_base import ConversationNestedEvaluationBase
        from ..models.conversation_nested_workflow_span_input_payload import ConversationNestedWorkflowSpanInputPayload
        from ..models.conversation_nested_workflow_span_output_payload import (
            ConversationNestedWorkflowSpanOutputPayload,
        )

        d = src_dict.copy()
        id = d.pop("id")

        name = d.pop("name")

        input_payload = ConversationNestedWorkflowSpanInputPayload.from_dict(d.pop("input_payload"))

        output_payload = ConversationNestedWorkflowSpanOutputPayload.from_dict(d.pop("output_payload"))

        started_at = isoparse(d.pop("started_at"))

        completed_at = isoparse(d.pop("completed_at"))

        evaluators = cast(list[str], d.pop("evaluators", UNSET))

        def _parse_parent_span_id(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        parent_span_id = _parse_parent_span_id(d.pop("parent_span_id", UNSET))

        evaluations = []
        _evaluations = d.pop("evaluations", UNSET)
        for evaluations_item_data in _evaluations or []:
            evaluations_item = ConversationNestedEvaluationBase.from_dict(evaluations_item_data)

            evaluations.append(evaluations_item)

        conversation_nested_workflow_span = cls(
            id=id,
            name=name,
            input_payload=input_payload,
            output_payload=output_payload,
            started_at=started_at,
            completed_at=completed_at,
            evaluators=evaluators,
            parent_span_id=parent_span_id,
            evaluations=evaluations,
        )

        conversation_nested_workflow_span.additional_properties = d
        return conversation_nested_workflow_span

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



"""Contains all the data models used in inputs/outputs"""

from .admin_evaluation_review_create_request import AdminEvaluationReviewCreateRequest
from .admin_evaluation_review_edit_request import AdminEvaluationReviewEditRequest
from .agent_target import AgentTarget
from .agent_target_create_request import AgentTargetCreateRequest
from .agent_target_revision import AgentTargetRevision
from .agent_target_revision_create_request import AgentTargetRevisionCreateRequest
from .agent_type import AgentType
from .agents_set_active_agent_target_response_agents_set_active_agent_target import (
    AgentsSetActiveAgentTargetResponseAgentsSetActiveAgentTarget,
)
from .conversation_nested import ConversationNested
from .conversation_nested_evaluation import ConversationNestedEvaluation
from .conversation_nested_evaluation_base import ConversationNestedEvaluationBase
from .conversation_nested_message import ConversationNestedMessage
from .conversation_nested_sentence import ConversationNestedSentence
from .conversation_nested_token_cost_base import ConversationNestedTokenCostBase
from .conversation_nested_turn import ConversationNestedTurn
from .conversation_nested_workflow_span import ConversationNestedWorkflowSpan
from .conversation_nested_workflow_span_input_payload import ConversationNestedWorkflowSpanInputPayload
from .conversation_nested_workflow_span_output_payload import ConversationNestedWorkflowSpanOutputPayload
from .conversation_nested_workflow_trace import ConversationNestedWorkflowTrace
from .conversation_turn_create_request import ConversationTurnCreateRequest
from .conversation_turn_create_response import ConversationTurnCreateResponse
from .create_test_run_request import CreateTestRunRequest
from .criteria import Criteria
from .criteria_instance_base import CriteriaInstanceBase
from .dataset import Dataset
from .dataset_create_request import DatasetCreateRequest
from .dataset_create_request_target_type import DatasetCreateRequestTargetType
from .dataset_item_create_item_request import DatasetItemCreateItemRequest
from .dataset_items_create_request import DatasetItemsCreateRequest
from .dataset_items_create_response import DatasetItemsCreateResponse
from .environment import Environment
from .error_codes import ErrorCodes
from .error_response import ErrorResponse
from .error_response_error import ErrorResponseError
from .evaluation_feedback import EvaluationFeedback
from .evaluation_feedback_create_request import EvaluationFeedbackCreateRequest
from .evaluation_feedback_create_request_feedback import EvaluationFeedbackCreateRequestFeedback
from .evaluation_feedback_feedback import EvaluationFeedbackFeedback
from .http_validation_error import HTTPValidationError
from .idp_org_create_request import IDPOrgCreateRequest
from .idp_org_user_add_request import IDPOrgUserAddRequest
from .idp_user_create_request import IDPUserCreateRequest
from .idp_user_update_request import IDPUserUpdateRequest
from .ingest_workflow_trace_response import IngestWorkflowTraceResponse
from .links import Links
from .metric import Metric
from .module import Module
from .module_metrics import ModuleMetrics
from .org import Org
from .org_create_request import OrgCreateRequest
from .test import Test
from .test_create_request import TestCreateRequest
from .test_create_request_documents_type_0 import TestCreateRequestDocumentsType0
from .test_create_request_entity_type import TestCreateRequestEntityType
from .test_create_request_metrics_config_type_0 import TestCreateRequestMetricsConfigType0
from .test_dataset import TestDataset
from .test_dataset_create_request import TestDatasetCreateRequest
from .test_documents_type_0 import TestDocumentsType0
from .test_entity_type import TestEntityType
from .test_metrics_config_type_0 import TestMetricsConfigType0
from .test_result_enum import TestResultEnum
from .test_run import TestRun
from .test_run_conversation_i_ds import TestRunConversationIDs
from .test_run_conversation_scores import TestRunConversationScores
from .test_run_metric_scores import TestRunMetricScores
from .test_run_metric_scores_criteria_failures_type_0 import TestRunMetricScoresCriteriaFailuresType0
from .test_run_results import TestRunResults
from .test_run_results_conversations import TestRunResultsConversations
from .test_run_results_conversations_metric_scores_type_0 import TestRunResultsConversationsMetricScoresType0
from .test_run_results_conversations_module_group_scores_type_0 import TestRunResultsConversationsModuleGroupScoresType0
from .test_run_results_conversations_module_scores_type_0 import TestRunResultsConversationsModuleScoresType0
from .test_run_results_metric_scores_type_0 import TestRunResultsMetricScoresType0
from .test_run_results_module_group_scores_type_0 import TestRunResultsModuleGroupScoresType0
from .test_run_results_module_scores_type_0 import TestRunResultsModuleScoresType0
from .test_run_status_update_request import TestRunStatusUpdateRequest
from .test_run_workflow_trace_i_ds import TestRunWorkflowTraceIDs
from .test_status_enum import TestStatusEnum
from .user import User
from .user_profile import UserProfile
from .v_evaluation_review_audit import VEvaluationReviewAudit
from .v_evaluation_review_state import VEvaluationReviewState
from .v_workflow import VWorkflow
from .validation_error import ValidationError
from .workflow_step_span_create_response import WorkflowStepSpanCreateResponse
from .workflow_step_span_create_response_input_payload_type_0 import WorkflowStepSpanCreateResponseInputPayloadType0

__all__ = (
    "AdminEvaluationReviewCreateRequest",
    "AdminEvaluationReviewEditRequest",
    "AgentsSetActiveAgentTargetResponseAgentsSetActiveAgentTarget",
    "AgentTarget",
    "AgentTargetCreateRequest",
    "AgentTargetRevision",
    "AgentTargetRevisionCreateRequest",
    "AgentType",
    "ConversationNested",
    "ConversationNestedEvaluation",
    "ConversationNestedEvaluationBase",
    "ConversationNestedMessage",
    "ConversationNestedSentence",
    "ConversationNestedTokenCostBase",
    "ConversationNestedTurn",
    "ConversationNestedWorkflowSpan",
    "ConversationNestedWorkflowSpanInputPayload",
    "ConversationNestedWorkflowSpanOutputPayload",
    "ConversationNestedWorkflowTrace",
    "ConversationTurnCreateRequest",
    "ConversationTurnCreateResponse",
    "CreateTestRunRequest",
    "Criteria",
    "CriteriaInstanceBase",
    "Dataset",
    "DatasetCreateRequest",
    "DatasetCreateRequestTargetType",
    "DatasetItemCreateItemRequest",
    "DatasetItemsCreateRequest",
    "DatasetItemsCreateResponse",
    "Environment",
    "ErrorCodes",
    "ErrorResponse",
    "ErrorResponseError",
    "EvaluationFeedback",
    "EvaluationFeedbackCreateRequest",
    "EvaluationFeedbackCreateRequestFeedback",
    "EvaluationFeedbackFeedback",
    "HTTPValidationError",
    "IDPOrgCreateRequest",
    "IDPOrgUserAddRequest",
    "IDPUserCreateRequest",
    "IDPUserUpdateRequest",
    "IngestWorkflowTraceResponse",
    "Links",
    "Metric",
    "Module",
    "ModuleMetrics",
    "Org",
    "OrgCreateRequest",
    "Test",
    "TestCreateRequest",
    "TestCreateRequestDocumentsType0",
    "TestCreateRequestEntityType",
    "TestCreateRequestMetricsConfigType0",
    "TestDataset",
    "TestDatasetCreateRequest",
    "TestDocumentsType0",
    "TestEntityType",
    "TestMetricsConfigType0",
    "TestResultEnum",
    "TestRun",
    "TestRunConversationIDs",
    "TestRunConversationScores",
    "TestRunMetricScores",
    "TestRunMetricScoresCriteriaFailuresType0",
    "TestRunResults",
    "TestRunResultsConversations",
    "TestRunResultsConversationsMetricScoresType0",
    "TestRunResultsConversationsModuleGroupScoresType0",
    "TestRunResultsConversationsModuleScoresType0",
    "TestRunResultsMetricScoresType0",
    "TestRunResultsModuleGroupScoresType0",
    "TestRunResultsModuleScoresType0",
    "TestRunStatusUpdateRequest",
    "TestRunWorkflowTraceIDs",
    "TestStatusEnum",
    "User",
    "UserProfile",
    "ValidationError",
    "VEvaluationReviewAudit",
    "VEvaluationReviewState",
    "VWorkflow",
    "WorkflowStepSpanCreateResponse",
    "WorkflowStepSpanCreateResponseInputPayloadType0",
)



from maihem.evaluators import MaihemEvaluator
from maihem.otel_client import Tracer, trace
from maihem.utils.utils import validate_attributes_testing, extract_ids_from_query
from maihem.clients import Maihem
from maihem.utils.registry import register_function
from typing import Optional, Callable, Any, Tuple, Dict
import inspect
from functools import wraps, lru_cache
import orjson as json
import os
import logging


def safe_serialize_input(data: Any) -> str:
    """
    Try to JSON-serialize the input.

    If JSON serialization fails and the data is a dict, try to iterate through each key/value:
    if an individual value fails, use its repr. For non-dict data fall back to repr.
    """
    try:
        return json.dumps(data)
    except Exception:
        if isinstance(data, dict):
            safe_data = {}
            for key, value in data.items():
                try:
                    # Try to safely serialize each value
                    safe_data[key] = json.loads(json.dumps(value))
                except Exception:
                    safe_data[key] = repr(value)
            try:
                return json.dumps(safe_data)
            except Exception:
                return repr(safe_data)
        return repr(data)


def safe_serialize_output(data: Any) -> str:
    """
    Try to JSON-serialize the output.

    If serialization fails, simply return a repr of the data.
    """
    try:
        return json.dumps(data)
    except Exception:
        return repr(data)


def _process_args_and_set_inputs(
    sig: inspect.Signature,
    args: Tuple[Any, ...],
    kwargs: Dict[str, Any],
    evaluator: Optional[MaihemEvaluator],
    span: Any,
    span_name: str,
    is_parent_span: bool = True,
) -> Tuple[Tuple[Any, ...], Dict[str, Any]]:
    """
    Bind and process the function arguments and update the span with the input payload.

    This function:
    - Binds the passed arguments using the cached signature.
    - Removes 'self' from the bound arguments.
    - Maps the inputs if an evaluator is provided.
    - Processes the 'query' parameter by extracting ids and converting
      the query to a "clean" version.
    - Sets the JSON-dumped representations of the payload on the span.

    Returns:
        Possibly modified (args, kwargs) after processing.
    """
    bound_args = sig.bind(*args, **kwargs)
    bound_args.apply_defaults()
    filtered_arguments = {k: v for k, v in bound_args.arguments.items() if k != "self"}

    # Use evaluator to map inputs if available.
    input_payload = (
        evaluator.map_inputs(**filtered_arguments) if evaluator else filtered_arguments
    )

    span.set_attribute("workflow_step_name", span_name)
    span.set_attribute("evaluator_name", evaluator.NAME if evaluator else None)
    span.set_attribute("input_payload", safe_serialize_input(input_payload))
    span.set_attribute("input_payload_raw", safe_serialize_input(filtered_arguments))

    client = Maihem()

    if is_parent_span:
        span.set_attribute("workflow_name", span_name)
        span.set_attribute("agent_target_id", client.target_agent_id)
        span.set_attribute("environment", client.environment)
        span.set_attribute("agent_target_revision_id", client.revision_id)

        if "query" in input_payload:
            clean_query, ids = extract_ids_from_query(input_payload["query"])
            for key, value in ids.items():
                span.set_attribute(key, value)

            # Replace query in args and kwargs if found.
            args = tuple(
                clean_query if arg == input_payload["query"] else arg for arg in args
            )
            if evaluator:
                arg_name = next(
                    (k for k, v in evaluator.input_mapping.items() if v == "query"),
                    None,
                )
                if arg_name and arg_name in kwargs:
                    kwargs[arg_name] = clean_query
            input_payload["query"] = clean_query
    else:
        span.set_attribute("entity_type", "workflow_step")

    return args, kwargs


def _set_output_attributes(
    span: Any, result: Any, evaluator: Optional[MaihemEvaluator]
) -> Any:
    """
    Map (if applicable) and set the output payload on the span.

    Returns:
        The mapped output.
    """
    output_payload = evaluator.map_outputs(result) if evaluator else result
    span.set_attribute("output_payload", safe_serialize_output(output_payload))
    span.set_attribute("output_payload_raw", safe_serialize_output(result))
    return output_payload


def workflow_step(
    name: Optional[str] = None,
    evaluator: Optional[MaihemEvaluator] = None,
) -> Callable:
    """
    Merged decorator that behaves as a workflow or a workflow step depending on context.

    - If no active parent span is present, it creates a top-level workflow span and (if provided)
      sets the agent_target_id attribute.
    - If an active parent span exists, it creates a child span (workflow step).
      It raises an error if the active span already has a parent (i.e. nested too deeply).

    Supports both asynchronous and synchronous functions.
    """

    def decorator(func: Callable) -> Callable:
        sig = inspect.signature(func)

        if inspect.iscoroutinefunction(func):

            @wraps(func)
            async def async_wrapper(*args: Any, **kwargs: Any) -> Any:
                parent_span = trace.get_current_span()
                is_parent_span = not parent_span.get_span_context().is_valid
                tracer = Tracer.get_instance().tracer
                span_name = name or func.__name__
                with tracer.start_as_current_span(span_name) as span:
                    new_args, new_kwargs = _process_args_and_set_inputs(
                        sig,
                        args,
                        kwargs,
                        evaluator,
                        span,
                        span_name,
                        is_parent_span,
                    )
                    result = await func(*new_args, **new_kwargs)
                    _set_output_attributes(span, result, evaluator)
                if not parent_span.get_span_context().is_valid:
                    tracer.span_processor.force_flush()
                return result

            return async_wrapper
        else:

            def sync_wrapper(*args: Any, **kwargs: Any) -> Any:
                parent_span = trace.get_current_span()
                is_parent_span = not parent_span.get_span_context().is_valid
                tracer = Tracer.get_instance().tracer
                span_name = name or func.__name__
                with tracer.start_as_current_span(span_name) as span:
                    new_args, new_kwargs = _process_args_and_set_inputs(
                        sig,
                        args,
                        kwargs,
                        evaluator,
                        span,
                        span_name,
                        is_parent_span,
                    )
                    result = func(*new_args, **new_kwargs)
                    _set_output_attributes(span, result, evaluator)
                    if not parent_span.get_span_context().is_valid:
                        tracer.span_processor.force_flush()
                    return result

            return sync_wrapper

    return decorator



from __future__ import annotations
import os
import loguru
from loguru import logger
from typing import TYPE_CHECKING
import sys

if TYPE_CHECKING:
    from loguru import Logger

extra_fields = [
    "request_id",
    "test_run_id",
    "conversation_id",
    "workflow_trace_id",
    "workflow_trace_raw_id",
    "workflow_span_id",
    "workflow_step_id",
]


class LoggerWrapper:
    def __init__(self, name: str, level="DEBUG", depth=1):
        self.name = name
        self.environment = os.getenv("ENVIRONMENT", "development")
        self._depth = depth

        # Remove default handler
        logger.remove()

        logger.add(
            sys.stdout,
            colorize=True,
            format=self.format_message,
            level=level,
            filter=self.add_extra_fields,
            backtrace=True,
            diagnose=False,
        )

    def format_message(self, record):
        # Add a relative pathname to 'extra', but do not overwrite existing keys
        record["extra"][
            "pathname"
        ] = f"{os.path.relpath(record['file'].path)}:{record['line']}"

        # Base format string
        base_format = (
            "<blue>{time:YYYY-MM-DD HH:mm:ss}</blue> - maihem - "
            "<level>{level}</level> - {message} - "
            "<dim>{extra[pathname]}</dim>"
        )

        # Add any additional 'extra' fields dynamically
        for key, value in record["extra"].items():
            if key == "pathname":  # Skip 'pathname' since it's already included
                continue
            if value:  # Only include non-empty values
                base_format += f" - <dim>{key}:{{extra[{key}]}}</dim>"

        # Include exception details if present
        if record["exception"]:
            base_format += "\n<red>{exception}</red>"

        base_format += "\n"
        return base_format

    def add_extra_fields(self, record):
        # record["extra"]["request_id"] = get_request_id()
        # record["extra"]["test_run_id"] = get_test_run_id()
        # record["extra"]["conversation_id"] = get_conversation_id()
        # record["extra"]["workflow_trace_id"] = get_workflow_trace_id()
        # record["extra"]["workflow_trace_raw_id"] = get_workflow_trace_raw_id()
        # record["extra"]["workflow_span_id"] = get_workflow_span_id()
        # record["extra"]["workflow_step_id"] = get_workflow_step_id()
        return True

    def get_logger(self) -> loguru.Logger:
        return logger


logger_instance = LoggerWrapper(name="maihem")
Logger = "loguru.Logger"


def get_logger() -> loguru.Logger:
    return logger_instance.get_logger()



import asyncio
from datetime import datetime
from pydantic import BaseModel
from enum import Enum
import os
from pydantic_extra_types.language_code import LanguageAlpha2
from typing import Optional, Tuple, List, Dict, Coroutine
import json
import importlib

from maihem.otel_client import Tracer
import maihem.shared.lib.errors as errors
from maihem.shared.lib.logger import get_logger

logger = get_logger()


class AgentType(str, Enum):
    MAIHEM = "maihem"
    TARGET = "target"


class TargetAgent(BaseModel):
    id: str
    created_at: datetime
    updated_at: datetime
    name: str
    label: Optional[str] = None
    role: Optional[str] = None
    description: Optional[str] = None
    industry: Optional[str] = None
    language: Optional[LanguageAlpha2] = "en"

    _wrapper_function: Optional[Coroutine] = None
    _wrapped_function_name: Optional[str] = None
    _workflow_name: Optional[str] = None
    document_paths: List[str] = []

    def set_wrapper_function(
        self,
        function_name: str,
        test_name: str,
        workflow_name: Optional[str] = None,
    ) -> None:
        """Dynamically imports and sets the wrapper function based on workflow name"""
        try:
            # Import the wrappers module
            wrappers = importlib.import_module(f"test_{test_name}.wrapper_functions")

            # Get the wrapper function name by prepending 'wrapper_' to workflow name
            wrapper_func_name = f"{function_name}_wrapper"

            # Get the function from the module
            if hasattr(wrappers, wrapper_func_name):
                self._wrapper_function = getattr(wrappers, wrapper_func_name)
                self._wrapped_function_name = function_name
                self._workflow_name = workflow_name
                logger.info(f"Wrapper function '{wrapper_func_name}' set successfully")
            else:
                raise AttributeError(
                    f"Wrapper function '{wrapper_func_name}' not found in wrappers.py"
                )

        except ImportError as e:
            raise ImportError(f"Could not import wrappers module: {str(e)}")
        except Exception as e:
            raise Exception(f"Error setting wrapper function: {str(e)}")

    def _test_wrapper_function(self, wrapper_function: Coroutine) -> bool:
        test_message = "Hi, it's the Maihem agent. Ready for testing?"
        logger.info("Testing wrapper function...")
        logger.info(f"Sending message to target agent: {test_message}")
        try:
            message, contexts = wrapper_function(str(datetime.now()), test_message, {})
            assert isinstance(message, str), "Response message must be a string"
            assert isinstance(contexts, list), "Contexts must be a list"
            for context in contexts:
                assert isinstance(
                    context, str
                ), "Each context in the list must be a string"

            return True
        except Exception as e:
            errors.raise_wrapper_function_error(
                logger=logger,
                message=f"Error testing target agent wrapper function: {e}",
            )

    def add_documents(self, documents: List[str]) -> None:
        for doc_path in documents:
            if not os.path.exists(doc_path):
                logger.warning(f"Document not found: {doc_path}")
                continue

            if not os.path.isfile(doc_path):
                logger.warning(f"Path is not a file: {doc_path}")
                continue

            file_size = os.path.getsize(doc_path)
            if file_size == 0:
                logger.warning(f"File is empty: {doc_path}")
                continue

            if file_size > 10 * 1024 * 1024:  # 10 MB limit
                logger.warning(f"File is too large (>10MB): {doc_path}")
                continue

            try:
                self.document_paths.append(doc_path)
                logger.info(f"Added document: {doc_path}")
            except Exception as e:
                logger.error(f"Error processing document {doc_path}: {str(e)}")

    def _call_workflow(
        self,
        conversation_id: str,
        conversation_message_id: str,
        message: Optional[str] = "",
        conversation_history: Dict = {},
        test_run_id: Optional[str] = None,
        target_agent_revision_id: Optional[str] = None,
        environment: Optional[str] = None,
    ) -> Tuple[str, List[str]]:
        if not self._wrapper_function:
            errors.raise_wrapper_function_error(
                logger=logger, message="Target agent wrapper function not set"
            )
        for retry in range(3):
            try:
                data = {
                    "query": message,
                    "maihem_ids": {
                        "conversation_id": conversation_id,
                        "conversation_message_id": conversation_message_id,
                        "test_run_id": test_run_id,
                        "agent_target_revision_id": target_agent_revision_id,
                        "environment": environment,
                    },
                }
                message_with_ids = json.dumps(data)

                kwargs = {
                    "conversation_id": conversation_id,
                    "user_input": message_with_ids or message,
                    "conversation_history": conversation_history,
                }

                response = asyncio.run(self._wrapper_function(**kwargs))
                return response
            except Exception as e:
                if retry < 2:
                    logger.warning(
                        f"Error sending message to target agent, retrying ({retry + 1}). Error: {str(e)}"
                    )
                else:
                    errors.raise_wrapper_function_error(
                        logger=logger,
                        message=f"Error sending message to target agent after 3 retries. Error: {str(e)}",
                    )

    def _call_step(
        self,
        interaction_id: str,
        target_agent_id: str,
        test_run_id: str,
        kwargs: Dict,
        target_agent_revision_id: Optional[str] = None,
        environment: Optional[str] = None,
    ) -> Tuple[str, List[str]]:
        if not self._wrapper_function:
            errors.raise_wrapper_function_error(
                logger=logger, message="Target agent wrapper function not set"
            )
        for retry in range(3):
            try:

                tracer = Tracer.get_instance().tracer
                # span_name = self._wrapped_function_name
                with tracer.start_as_current_span(self._workflow_name) as span:
                    span.set_attribute("workflow_trace_id", interaction_id)
                    span.set_attribute("test_run_id", test_run_id)
                    span.set_attribute("agent_target_id", target_agent_id)
                    span.set_attribute("workflow_name", self._workflow_name)
                    span.set_attribute("workflow_step_name", self._workflow_name)
                    if target_agent_revision_id:
                        span.set_attribute(
                            "agent_target_revision_id", target_agent_revision_id
                        )
                    if environment:
                        span.set_attribute("environment", environment)

                    # Call the step wrapper function
                    response = asyncio.run(self._wrapper_function(**kwargs))
                tracer.span_processor.force_flush()
                return response
            except Exception as e:
                if retry < 2:
                    logger.warning(
                        f"Error sending message to target agent, retrying ({retry + 1}). Error: {str(e)}"
                    )
                else:
                    errors.raise_wrapper_function_error(
                        logger=logger,
                        message=f"Error sending message to target agent after 3 retries. Error: {str(e)}",
                    )



from enum import Enum


class Environment(str, Enum):
    DEVELOPMENT = "development"
    PRODUCTION = "production"
    STAGING = "staging"
    TEST = "test"

    def __str__(self) -> str:
        return str(self.value)



from typing import Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="AdminEvaluationReviewEditRequest")


@_attrs_define
class AdminEvaluationReviewEditRequest:
    """
    Attributes:
        result (str):
        entity_id (str):
        entity_type (str):
        evaluation_id (Union[None, Unset, str]):
        score (Union[None, Unset, int]):
        explanation (Union[None, Unset, str]):
    """

    result: str
    entity_id: str
    entity_type: str
    evaluation_id: Union[None, Unset, str] = UNSET
    score: Union[None, Unset, int] = UNSET
    explanation: Union[None, Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        result = self.result

        entity_id = self.entity_id

        entity_type = self.entity_type

        evaluation_id: Union[None, Unset, str]
        if isinstance(self.evaluation_id, Unset):
            evaluation_id = UNSET
        else:
            evaluation_id = self.evaluation_id

        score: Union[None, Unset, int]
        if isinstance(self.score, Unset):
            score = UNSET
        else:
            score = self.score

        explanation: Union[None, Unset, str]
        if isinstance(self.explanation, Unset):
            explanation = UNSET
        else:
            explanation = self.explanation

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "result": result,
                "entity_id": entity_id,
                "entity_type": entity_type,
            }
        )
        if evaluation_id is not UNSET:
            field_dict["evaluation_id"] = evaluation_id
        if score is not UNSET:
            field_dict["score"] = score
        if explanation is not UNSET:
            field_dict["explanation"] = explanation

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        d = src_dict.copy()
        result = d.pop("result")

        entity_id = d.pop("entity_id")

        entity_type = d.pop("entity_type")

        def _parse_evaluation_id(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        evaluation_id = _parse_evaluation_id(d.pop("evaluation_id", UNSET))

        def _parse_score(data: object) -> Union[None, Unset, int]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, int], data)

        score = _parse_score(d.pop("score", UNSET))

        def _parse_explanation(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        explanation = _parse_explanation(d.pop("explanation", UNSET))

        admin_evaluation_review_edit_request = cls(
            result=result,
            entity_id=entity_id,
            entity_type=entity_type,
            evaluation_id=evaluation_id,
            score=score,
            explanation=explanation,
        )

        admin_evaluation_review_edit_request.additional_properties = d
        return admin_evaluation_review_edit_request

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from abc import ABC, abstractmethod, ABCMeta
from typing import (
    Any,
    Dict,
    Optional,
    Mapping,
    ClassVar,
    Set,
    Callable,
    get_type_hints,
    Type,
)
from pydantic import BaseModel  # Install via: pip install pydantic
import textwrap


# Optional: define a custom exception for InputMapping errors
class InputMappingError(TypeError):
    """Exception raised when InputMapping validation fails."""

    pass


# Use Pydantic for robust input mapping validation.
class InputMapping(BaseModel):
    """
    Maps standard input names to your function's parameter names.
    The attribute names represent the standard interface, and the values
    are your function's parameter names.

    Pydantic automatically validates the types based on the annotations.
    """

    def to_dict(self) -> Dict[str, str]:
        """
        Returns a mapping from your function's parameter names to the standard names.
        The reverse mapping is defined as: the value stored in the instance (i.e. your
        function parameter) maps to the field name (i.e. the standardized input).
        """
        # Use model_fields instead of __fields__ (deprecated in recent Pydantic versions)
        return {getattr(self, field): field for field in self.model_fields}

    class Config:
        # Disallow extra fields so that only annotated fields are allowed.
        extra = "forbid"


# Global registry for storing Evaluator classes
EVALUATOR_REGISTRY: Dict[str, Type["MaihemEvaluator"]] = {}


class EvaluatorMeta(ABCMeta):
    """
    Metaclass for MaihemEvaluator that automatically sets the
    EXPECTED_INPUTS from the inner Inputs class annotations.
    """

    def __init__(cls, name, bases, namespace):
        super().__init__(name, bases, namespace)
        inputs_cls = getattr(cls, "Inputs", None)
        if inputs_cls:
            # Using get_type_hints ensures that forward references and other niceties are handled.
            cls.EXPECTED_INPUTS = set(get_type_hints(inputs_cls).keys())
        else:
            cls.EXPECTED_INPUTS = set()
        # Register evaluator classes with a non-empty NAME
        if cls.NAME:
            if cls.NAME in EVALUATOR_REGISTRY:
                raise ValueError(f"Duplicate evaluator name: {cls.NAME}")
            EVALUATOR_REGISTRY[cls.NAME] = cls


class MaihemEvaluator(ABC, metaclass=EvaluatorMeta):
    # Connector name for identification must be defined in each subclass.
    NAME: ClassVar[str] = ""
    EXPECTED_INPUTS: Set[str] = set()  # Default for static type-checking

    def __init__(
        self,
        inputs: Optional[InputMapping] = None,
        output_fn: Optional[Callable[[Any], Any]] = None,
    ) -> None:
        """
        Initialize connector with optional input mapping and output transformation.

        Args:
            inputs: Maps your function's parameter names to our standard names. Use
                get_expected_inputs() to see what standard inputs are needed.
            output_fn: Optional function to transform the decorated function's output
                before standard output mapping.
        """
        if not self.NAME:
            raise ValueError(f"Connector {self.__class__.__name__} must define NAME")

        self.input_mapping: Dict[str, str] = inputs.to_dict() if inputs else {}
        self.output_fn = output_fn
        self._validate_mappings()

    def _validate_mappings(self) -> None:
        """
        Validates that all provided input mappings are among the expected inputs.
        """
        if self.input_mapping:
            # The mapping's values are the standard input names.
            mapped_inputs = set(self.input_mapping.values())
            invalid_inputs = mapped_inputs - self.EXPECTED_INPUTS
            if invalid_inputs:
                raise ValueError(
                    f"Invalid input mappings: {invalid_inputs}. "
                    f"Expected inputs are: {self.EXPECTED_INPUTS}"
                )

    @classmethod
    def get_expected_inputs(cls) -> Set[str]:
        """
        Returns the set of standard input names that this connector expects.
        """
        return cls.EXPECTED_INPUTS

    @classmethod
    def create_with_mapping_help(cls) -> None:
        """
        Prints help text showing what mappings this connector needs.
        """
        print(f"\nConnector: {cls.__name__}\n")
        print("Expected Inputs:")
        for input_name in sorted(cls.EXPECTED_INPUTS):
            print(f"  - {input_name}")
        print("\nExample usage:")
        example_inputs = {f"your_{name}": name for name in sorted(cls.EXPECTED_INPUTS)}
        print(
            f"""
@maihem(
    connector={cls.__name__}(
        inputs=InputMapping(
            {", ".join(f"{k}='{v}'" for k, v in example_inputs.items())}
        )
    )
)
"""
        )

    def map_inputs(self, **kwargs: Any) -> Dict[str, Any]:
        """
        Map function inputs to standardized names using the provided input mapping.
        Only inputs with an explicit mapping are included.
        """
        return {
            self.input_mapping[k]: v
            for k, v in kwargs.items()
            if k in self.input_mapping
        }

    def map_outputs(self, result: Any) -> Dict[str, Any]:
        """
        Transform the function output to a standardized format.
        Optionally applies the defined output transformation.
        """
        if self.output_fn:
            result = self.output_fn(result)
        return self._transform_output(result)

    @abstractmethod
    def _transform_output(self, result: Any) -> Dict[str, Any]:
        """
        Transform the return value to a standardized format. Must be implemented
        by subclasses.
        """
        pass

    def _generate_function_wrapper(self, function_name: str) -> str:
        """
        Generate a wrapper code snippet for the connector function.
        (In practice, you might return a callable or write this to a file.)
        """
        # Get type hints from the inner Inputs class.
        inputs_hints = get_type_hints(self.Inputs)
        args_str = ", ".join(
            f"{k}: {v.__name__ if hasattr(v, '__name__') else str(v)}"
            for k, v in inputs_hints.items()
        )
        # Create required inputs with proper indentation handling
        required_inputs = [
            f"    - {k}: {v.__name__ if hasattr(v, '__name__') else str(v)}"
            for k, v in inputs_hints.items()
        ]
        params = ", ".join(inputs_hints.keys())

        # Build the wrapper code with explicit newlines and indentation
        wrapper_code = [
            f"###### {self.NAME} Wrapper ######",
            "",
            "# Add imports here",
            f"# import my_{self.NAME}",
            "",
            "",
            f"def {function_name}_wrapper({args_str}):",
            f'    """Wrapper for {self.NAME} step.',
            "",
            "    Required Inputs:",
        ]
        wrapper_code.extend(required_inputs)
        wrapper_code.extend(
            [
                '    """',
                "",
                "    ##### YOUR CODE HERE #####",
                f"    # my_{self.NAME}({params})",
            ]
        )

        return "\n".join(wrapper_code)


class MaihemConversational(MaihemEvaluator):
    """Conversational connector.

    Output Format:
        answer: str  # The answer to the question
    """

    NAME = "conversational"

    class Inputs(InputMapping):
        """
        Input mapping for the Question Answering connector.

        Attributes:
            query (str): The name of the argument that contains the query as a string.
        """

        query: str

    def _transform_output(self, result: str) -> Dict[str, Any]:
        return {"answer": result}


class MaihemIntent(MaihemEvaluator):
    """Intent recognition connector.

    Output Format:
        intent: str  # The classified intent
    """

    NAME = "intent_classification"

    class Inputs(InputMapping):
        """
        Input mapping for the Intent Recognition connector.

        Attributes:
            query (str): The name of the argument that contains the query as a string.
        """

        query: str

    def _transform_output(self, result: str) -> Dict[str, Any]:
        return {"intent": result}


class MaihemNER(MaihemEvaluator):
    """Named Entity Recognition connector.

    Output Format:
        entities: list[str]  # List of extracted entities
    """

    NAME = "named_entity_recognition"

    class Inputs(InputMapping):
        """
        Input mapping for the Named Entity Recognition connector.

        Attributes:
            query (str): The name of the argument that contains the query as a string.
        """

        query: str

    def _transform_output(self, result: list[str]) -> Dict[str, Any]:
        return {"entities": result}


class MaihemQueryRewriting(MaihemEvaluator):
    """Query rewriting connector.

    Output Format:
        rewritten_query: str  # The rephrased query
    """

    NAME = "query_rewriting"

    class Inputs(InputMapping):
        """
        Input mapping for the Query Rephrasing connector.

        Attributes:
            query (str): The name of the argument that contains the query as a string.
        """

        query: str

    def _transform_output(self, result: str) -> Dict[str, Any]:
        return {"rewritten_query": result}


class MaihemSearch(MaihemEvaluator):
    """Search connector.

    Output Format:
        documents: list[str]  # Retrieved text chunks
    """

    NAME = "search"

    class Inputs(InputMapping):
        """
        Input mapping for the Search connector.

        Attributes:
            query (str): The name of the argument that contains the query as a string.
        """

        query: str

    def _transform_output(self, result: list[str]) -> Dict[str, Any]:
        return {"documents": result}


class MaihemReranking(MaihemEvaluator):
    """Document reranking connector.

    Output Format:
        reranked_documents: list[str]  # Reranked documents
    """

    NAME = "reranking"

    class Inputs(InputMapping):
        """
        Input mapping for the Document Reranking connector.

        Attributes:
            query (str): The name of the argument that contains the query as a string.
            documents (str): The name of the argument that contains the documents as a list of strings.
        """

        query: str
        documents: str

    def _transform_output(self, result: list[str]) -> Dict[str, Any]:
        return {"reranked_documents": result}


class MaihemFiltering(MaihemEvaluator):
    """Document filtering connector.

    Output Format:
        filtered_documents: list[str]  # Filtered documents
    """

    NAME = "filtering"

    class Inputs(InputMapping):
        """
        Input mapping for the Document Filtering connector.

        Attributes:
            query (str): The name of the argument that contains the query as a string.
            documents (str): The name of the argument that contains the documents as a list of strings.
        """

        query: str
        documents: str

    def _transform_output(self, result: list[str]) -> Dict[str, Any]:
        return {"filtered_documents": result}


class MaihemSummarization(MaihemEvaluator):
    """Summarization connector.

    Output Format:
        summary: str  # The summary
    """

    NAME = "summarization"

    class Inputs(InputMapping):
        """
        Input mapping for the Summarization connector.

        Attributes:
            texts (list[str]): The name of the argument that contains the texts as a list of strings.
        """

        texts: str

    def _transform_output(self, result: list[str]) -> Dict[str, Any]:
        return {"summarized_texts": result}


class MaihemFinalAnswer(MaihemEvaluator):
    """Final answer generation connector.

    Output Format:
        answer: str  # The final answer
    """

    NAME = "answer_generation"

    class Inputs(InputMapping):
        """
        Input mapping for the Final Answer Generation connector.

        Attributes:
            query (str): The name of the argument that contains the query as a string.
            documents (str): The name of the argument that contains the documents as a list of strings.
        """

        query: str
        documents: str

    def _transform_output(self, result: str) -> Dict[str, Any]:
        return {"answer": result}



from typing import Any, Callable, Optional

function_registry = {}


def register_function(function: Callable, function_name: Optional[str] = None) -> None:
    function_name_dict = function_name or function.__name__
    function_registry[function_name_dict] = function


def call_function(function_name: str, *args, **kwargs) -> Any:
    return function_registry[function_name](*args, **kwargs)


def get_function(function_name: str) -> Callable:
    return function_registry[function_name]



from typing import TYPE_CHECKING, Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.dataset_item_create_item_request import DatasetItemCreateItemRequest


T = TypeVar("T", bound="DatasetItemsCreateRequest")


@_attrs_define
class DatasetItemsCreateRequest:
    """
    Attributes:
        items (Union[Unset, list['DatasetItemCreateItemRequest']]):
    """

    items: Union[Unset, list["DatasetItemCreateItemRequest"]] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        items: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.items, Unset):
            items = []
            for items_item_data in self.items:
                items_item = items_item_data.to_dict()
                items.append(items_item)

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if items is not UNSET:
            field_dict["items"] = items

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        from ..models.dataset_item_create_item_request import DatasetItemCreateItemRequest

        d = src_dict.copy()
        items = []
        _items = d.pop("items", UNSET)
        for items_item_data in _items or []:
            items_item = DatasetItemCreateItemRequest.from_dict(items_item_data)

            items.append(items_item)

        dataset_items_create_request = cls(
            items=items,
        )

        dataset_items_create_request.additional_properties = d
        return dataset_items_create_request

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.environment import Environment
from ..types import UNSET, Unset

T = TypeVar("T", bound="CreateTestRunRequest")


@_attrs_define
class CreateTestRunRequest:
    """
    Attributes:
        name (str):
        label (Union[None, Unset, str]):
        environment (Union[Environment, None, Unset]):
        agent_target_revision_id (Union[None, Unset, str]):
    """

    name: str
    label: Union[None, Unset, str] = UNSET
    environment: Union[Environment, None, Unset] = UNSET
    agent_target_revision_id: Union[None, Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        name = self.name

        label: Union[None, Unset, str]
        if isinstance(self.label, Unset):
            label = UNSET
        else:
            label = self.label

        environment: Union[None, Unset, str]
        if isinstance(self.environment, Unset):
            environment = UNSET
        elif isinstance(self.environment, Environment):
            environment = self.environment.value
        else:
            environment = self.environment

        agent_target_revision_id: Union[None, Unset, str]
        if isinstance(self.agent_target_revision_id, Unset):
            agent_target_revision_id = UNSET
        else:
            agent_target_revision_id = self.agent_target_revision_id

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "name": name,
            }
        )
        if label is not UNSET:
            field_dict["label"] = label
        if environment is not UNSET:
            field_dict["environment"] = environment
        if agent_target_revision_id is not UNSET:
            field_dict["agent_target_revision_id"] = agent_target_revision_id

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        d = src_dict.copy()
        name = d.pop("name")

        def _parse_label(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        label = _parse_label(d.pop("label", UNSET))

        def _parse_environment(data: object) -> Union[Environment, None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, str):
                    raise TypeError()
                environment_type_0 = Environment(data)

                return environment_type_0
            except:  # noqa: E722
                pass
            return cast(Union[Environment, None, Unset], data)

        environment = _parse_environment(d.pop("environment", UNSET))

        def _parse_agent_target_revision_id(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        agent_target_revision_id = _parse_agent_target_revision_id(d.pop("agent_target_revision_id", UNSET))

        create_test_run_request = cls(
            name=name,
            label=label,
            environment=environment,
            agent_target_revision_id=agent_target_revision_id,
        )

        create_test_run_request.additional_properties = d
        return create_test_run_request

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="DatasetItemCreateItemRequest")


@_attrs_define
class DatasetItemCreateItemRequest:
    """
    Attributes:
        input_payload (str):
        external_id (Union[None, Unset, str]):
        output_payload_expected (Union[None, Unset, str]):
        conversation_history (Union[None, Unset, str]):
    """

    input_payload: str
    external_id: Union[None, Unset, str] = UNSET
    output_payload_expected: Union[None, Unset, str] = UNSET
    conversation_history: Union[None, Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        input_payload = self.input_payload

        external_id: Union[None, Unset, str]
        if isinstance(self.external_id, Unset):
            external_id = UNSET
        else:
            external_id = self.external_id

        output_payload_expected: Union[None, Unset, str]
        if isinstance(self.output_payload_expected, Unset):
            output_payload_expected = UNSET
        else:
            output_payload_expected = self.output_payload_expected

        conversation_history: Union[None, Unset, str]
        if isinstance(self.conversation_history, Unset):
            conversation_history = UNSET
        else:
            conversation_history = self.conversation_history

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "input_payload": input_payload,
            }
        )
        if external_id is not UNSET:
            field_dict["external_id"] = external_id
        if output_payload_expected is not UNSET:
            field_dict["output_payload_expected"] = output_payload_expected
        if conversation_history is not UNSET:
            field_dict["conversation_history"] = conversation_history

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        d = src_dict.copy()
        input_payload = d.pop("input_payload")

        def _parse_external_id(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        external_id = _parse_external_id(d.pop("external_id", UNSET))

        def _parse_output_payload_expected(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        output_payload_expected = _parse_output_payload_expected(d.pop("output_payload_expected", UNSET))

        def _parse_conversation_history(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        conversation_history = _parse_conversation_history(d.pop("conversation_history", UNSET))

        dataset_item_create_item_request = cls(
            input_payload=input_payload,
            external_id=external_id,
            output_payload_expected=output_payload_expected,
            conversation_history=conversation_history,
        )

        dataset_item_create_item_request.additional_properties = d
        return dataset_item_create_item_request

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from contextvars import ContextVar

request_id_context: ContextVar[str] = ContextVar("request_id", default="")
conversation_id_context: ContextVar[str] = ContextVar("conversation_id", default="")
test_run_id_context: ContextVar[str] = ContextVar("test_run_id", default="")
workflow_trace_raw_id_context: ContextVar[str] = ContextVar(
    "workflow_trace_raw_id", default=""
)
workflow_trace_id_context: ContextVar[str] = ContextVar("workflow_trace_id", default="")
workflow_span_id_context: ContextVar[str] = ContextVar("workflow_span_id", default="")
workflow_step_id_context: ContextVar[str] = ContextVar("workflow_step_id", default="")


def get_request_id() -> str:
    return request_id_context.get()


def set_request_id(request_id: str):
    request_id_context.set(request_id)


def get_conversation_id() -> str:
    return conversation_id_context.get()


def set_conversation_id(conversation_id: str):
    conversation_id_context.set(conversation_id)


def get_test_run_id() -> str:
    return test_run_id_context.get()


def set_test_run_id(test_id: str):
    test_run_id_context.set(test_id)


def get_workflow_trace_raw_id() -> str:
    return workflow_trace_raw_id_context.get()


def set_workflow_trace_raw_id(workflow_trace_raw_id: str):
    workflow_trace_raw_id_context.set(workflow_trace_raw_id)


def get_workflow_trace_id() -> str:
    return workflow_trace_id_context.get()


def set_workflow_trace_id(workflow_trace_id: str):
    workflow_trace_id_context.set(workflow_trace_id)


def get_workflow_span_id() -> str:
    return workflow_span_id_context.get()


def set_workflow_span_id(workflow_span_id: str):
    workflow_span_id_context.set(workflow_span_id)


def get_workflow_step_id() -> str:
    return workflow_step_id_context.get()


def set_workflow_step_id(workflow_step_id: str):
    workflow_step_id_context.set(workflow_step_id)



from typing import TYPE_CHECKING, Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.workflow_step_span_create_response_input_payload_type_0 import (
        WorkflowStepSpanCreateResponseInputPayloadType0,
    )


T = TypeVar("T", bound="WorkflowStepSpanCreateResponse")


@_attrs_define
class WorkflowStepSpanCreateResponse:
    """
    Attributes:
        input_payload (Union['WorkflowStepSpanCreateResponseInputPayloadType0', None, Unset]):
    """

    input_payload: Union["WorkflowStepSpanCreateResponseInputPayloadType0", None, Unset] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        from ..models.workflow_step_span_create_response_input_payload_type_0 import (
            WorkflowStepSpanCreateResponseInputPayloadType0,
        )

        input_payload: Union[None, Unset, dict[str, Any]]
        if isinstance(self.input_payload, Unset):
            input_payload = UNSET
        elif isinstance(self.input_payload, WorkflowStepSpanCreateResponseInputPayloadType0):
            input_payload = self.input_payload.to_dict()
        else:
            input_payload = self.input_payload

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if input_payload is not UNSET:
            field_dict["input_payload"] = input_payload

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        from ..models.workflow_step_span_create_response_input_payload_type_0 import (
            WorkflowStepSpanCreateResponseInputPayloadType0,
        )

        d = src_dict.copy()

        def _parse_input_payload(data: object) -> Union["WorkflowStepSpanCreateResponseInputPayloadType0", None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                input_payload_type_0 = WorkflowStepSpanCreateResponseInputPayloadType0.from_dict(data)

                return input_payload_type_0
            except:  # noqa: E722
                pass
            return cast(Union["WorkflowStepSpanCreateResponseInputPayloadType0", None, Unset], data)

        input_payload = _parse_input_payload(d.pop("input_payload", UNSET))

        workflow_step_span_create_response = cls(
            input_payload=input_payload,
        )

        workflow_step_span_create_response.additional_properties = d
        return workflow_step_span_create_response

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import datetime
from typing import TYPE_CHECKING, Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.conversation_nested_workflow_span import ConversationNestedWorkflowSpan


T = TypeVar("T", bound="ConversationNestedWorkflowTrace")


@_attrs_define
class ConversationNestedWorkflowTrace:
    """
    Attributes:
        id (str):
        workflow_id (str):
        started_at (datetime.datetime):
        completed_at (datetime.datetime):
        created_at (Union[None, Unset, datetime.datetime]):
        updated_at (Union[None, Unset, datetime.datetime]):
        workflow_spans (Union[Unset, list['ConversationNestedWorkflowSpan']]):
    """

    id: str
    workflow_id: str
    started_at: datetime.datetime
    completed_at: datetime.datetime
    created_at: Union[None, Unset, datetime.datetime] = UNSET
    updated_at: Union[None, Unset, datetime.datetime] = UNSET
    workflow_spans: Union[Unset, list["ConversationNestedWorkflowSpan"]] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        id = self.id

        workflow_id = self.workflow_id

        started_at = self.started_at.isoformat()

        completed_at = self.completed_at.isoformat()

        created_at: Union[None, Unset, str]
        if isinstance(self.created_at, Unset):
            created_at = UNSET
        elif isinstance(self.created_at, datetime.datetime):
            created_at = self.created_at.isoformat()
        else:
            created_at = self.created_at

        updated_at: Union[None, Unset, str]
        if isinstance(self.updated_at, Unset):
            updated_at = UNSET
        elif isinstance(self.updated_at, datetime.datetime):
            updated_at = self.updated_at.isoformat()
        else:
            updated_at = self.updated_at

        workflow_spans: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.workflow_spans, Unset):
            workflow_spans = []
            for workflow_spans_item_data in self.workflow_spans:
                workflow_spans_item = workflow_spans_item_data.to_dict()
                workflow_spans.append(workflow_spans_item)

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "id": id,
                "workflow_id": workflow_id,
                "started_at": started_at,
                "completed_at": completed_at,
            }
        )
        if created_at is not UNSET:
            field_dict["created_at"] = created_at
        if updated_at is not UNSET:
            field_dict["updated_at"] = updated_at
        if workflow_spans is not UNSET:
            field_dict["workflow_spans"] = workflow_spans

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        from ..models.conversation_nested_workflow_span import ConversationNestedWorkflowSpan

        d = src_dict.copy()
        id = d.pop("id")

        workflow_id = d.pop("workflow_id")

        started_at = isoparse(d.pop("started_at"))

        completed_at = isoparse(d.pop("completed_at"))

        def _parse_created_at(data: object) -> Union[None, Unset, datetime.datetime]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, str):
                    raise TypeError()
                created_at_type_0 = isoparse(data)

                return created_at_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, Unset, datetime.datetime], data)

        created_at = _parse_created_at(d.pop("created_at", UNSET))

        def _parse_updated_at(data: object) -> Union[None, Unset, datetime.datetime]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, str):
                    raise TypeError()
                updated_at_type_0 = isoparse(data)

                return updated_at_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, Unset, datetime.datetime], data)

        updated_at = _parse_updated_at(d.pop("updated_at", UNSET))

        workflow_spans = []
        _workflow_spans = d.pop("workflow_spans", UNSET)
        for workflow_spans_item_data in _workflow_spans or []:
            workflow_spans_item = ConversationNestedWorkflowSpan.from_dict(workflow_spans_item_data)

            workflow_spans.append(workflow_spans_item)

        conversation_nested_workflow_trace = cls(
            id=id,
            workflow_id=workflow_id,
            started_at=started_at,
            completed_at=completed_at,
            created_at=created_at,
            updated_at=updated_at,
            workflow_spans=workflow_spans,
        )

        conversation_nested_workflow_trace.additional_properties = d
        return conversation_nested_workflow_trace

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import datetime
from typing import Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

T = TypeVar("T", bound="VWorkflow")


@_attrs_define
class VWorkflow:
    """
    Attributes:
        id (str):
        created_at (datetime.datetime):
        updated_at (datetime.datetime):
        name (str):
        label (Union[None, str]):
        org_id (str):
        agent_target_id (str):
        workflow_steps (Any):
    """

    id: str
    created_at: datetime.datetime
    updated_at: datetime.datetime
    name: str
    label: Union[None, str]
    org_id: str
    agent_target_id: str
    workflow_steps: Any
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        id = self.id

        created_at = self.created_at.isoformat()

        updated_at = self.updated_at.isoformat()

        name = self.name

        label: Union[None, str]
        label = self.label

        org_id = self.org_id

        agent_target_id = self.agent_target_id

        workflow_steps = self.workflow_steps

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "id": id,
                "created_at": created_at,
                "updated_at": updated_at,
                "name": name,
                "label": label,
                "org_id": org_id,
                "agent_target_id": agent_target_id,
                "workflow_steps": workflow_steps,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: dict[str, Any]) -> T:
        d = src_dict.copy()
        id = d.pop("id")

        created_at = isoparse(d.pop("created_at"))

        updated_at = isoparse(d.pop("updated_at"))

        name = d.pop("name")

        def _parse_label(data: object) -> Union[None, str]:
            if data is None:
                return data
            return cast(Union[None, str], data)

        label = _parse_label(d.pop("label"))

        org_id = d.pop("org_id")

        agent_target_id = d.pop("agent_target_id")

        workflow_steps = d.pop("workflow_steps")

        v_workflow = cls(
            id=id,
            created_at=created_at,
            updated_at=updated_at,
            name=name,
            label=label,
            org_id=org_id,
            agent_target_id=agent_target_id,
            workflow_steps=workflow_steps,
        )

        v_workflow.additional_properties = d
        return v_workflow

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from enum import Enum
from typing import Dict, Optional, Any
from pydantic import ValidationError, BaseModel
from .request_context import get_request_id
from .logger import Logger


class ErrorCodes(str, Enum):
    ERR_NOT_FOUND = "err_not_found"
    ERR_WORKFLOW_PROCESSING = "err_workflow_processing"
    ERR_INTERNAL_SERVER = "err_internal_server"
    ERR_REQUEST_VALIDATION = "err_request_validation"
    ERR_SCHEMA_VALIDATION = "err_schema_validation"
    ERR_DATA_INTEGRITY = "err_data_integrity"
    ERR_AUTHENTICATION_FAILED = "err_authentication_failed"
    ERR_REQUEST_TIMEOUT = "err_request_timeout"
    ERR_NOT_IMPLEMENTED = "err_not_implemented"
    ERR_WRAPPER_FUNCTION = "err_wrapper_function"
    ERR_CONFIG_FILE = "err_config_file"


class ErrorResponseError(BaseModel):
    code: ErrorCodes
    message: str
    detail: Optional[Any] = None


class ErrorResponse(BaseModel):
    error: ErrorResponseError
    request_id: Optional[str] = None


error_response_base = {
    400: {"model": ErrorResponse},
    409: {"model": ErrorResponse},
    422: {"model": ErrorResponse},
    500: {"model": ErrorResponse},
    504: {"model": ErrorResponse},
}


class BaseError(Exception):
    def __init__(
        self,
        status_code=500,
        code=ErrorCodes.ERR_INTERNAL_SERVER,
        message="Internal server error",
        detail=None,
    ):
        if detail is None:
            detail = {}
        self.status_code = status_code
        self.code = code
        self.message = message
        self.detail = detail
        super().__init__(self.message)

    def to_dict(self) -> Dict:
        error_body = ErrorResponseError(
            code=self.code, message=self.message, detail=self.detail
        )
        error = ErrorResponse(error=error_body, request_id=get_request_id())

        return error.model_dump()


class AuthenticationFailedError(BaseError):
    def __init__(
        self,
        message: str = "Authentication failed",
    ):
        super().__init__(
            status_code=401,
            code=ErrorCodes.ERR_AUTHENTICATION_FAILED,
            message=message,
            detail=None,
        )


class NotFoundError(BaseError):
    def __init__(self, entity_type: str, entity_key: str = None):
        if entity_key:
            message = f"{entity_type} ({entity_key}) not found"
        else:
            message = f"{entity_type} not found"
        super().__init__(
            status_code=404,
            code=ErrorCodes.ERR_NOT_FOUND,
            message=message,
            detail={"entity_type": entity_type, "entity_key": entity_key},
        )


class SchemaValidationError(BaseError):
    def __init__(self, message=None, detail=None):
        super().__init__(
            status_code=400,
            code=ErrorCodes.ERR_SCHEMA_VALIDATION,
            message="Schema validation error" + (f": {message}" if message else ""),
            detail=detail,
        )


class DataIntegrityError(BaseError):
    def __init__(self, entity_type: str, entity_key: str, detail=None):
        super().__init__(
            status_code=409,
            code=ErrorCodes.ERR_DATA_INTEGRITY,
            message=f"{entity_type} ({entity_key}) already exists",
            detail=detail,
        )


class RequestTimeoutError(BaseError):
    def __init__(self, message: str = "Request timeout", detail=None):
        super().__init__(
            status_code=504,
            code=ErrorCodes.ERR_REQUEST_TIMEOUT,
            message=f"{message}",
            detail=detail,
        )


class RequestValidationCustomError(BaseError):
    def __init__(self, message: str = "Request validation error", detail=None):
        super().__init__(
            status_code=400,
            code=ErrorCodes.ERR_REQUEST_VALIDATION,
            message=f"{message}",
            detail=detail,
        )


class WorkflowProcessingError(BaseError):
    def __init__(self, message: str = "Error processing workflow", detail=None):
        super().__init__(
            status_code=500,
            code=ErrorCodes.ERR_WORKFLOW_PROCESSING,
            message=f"Error processing workflow: {message}",
            detail=detail,
        )


class NotImplementError(BaseError):
    def __init__(self, message: str = "Not implemented", detail=None):
        super().__init__(
            status_code=501,
            code=ErrorCodes.ERR_NOT_IMPLEMENTED,
            message=f"{message}",
            detail=detail,
        )


class InternalServerError(BaseError):
    def __init__(self, message: str = "Internal server error", detail=None):
        super().__init__(
            status_code=500,
            code=ErrorCodes.ERR_INTERNAL_SERVER,
            message=f"{message}",
            detail=detail,
        )


class WrapperFunctionError(BaseError):
    def __init__(self, message: str = "Wrapper function error", detail=None):
        super().__init__(
            status_code=500,
            code=ErrorCodes.ERR_WRAPPER_FUNCTION,
            message=f"{message}",
            detail=detail,
        )


class ConfigFileError(BaseError):
    def __init__(self, message: str = "Config file error", detail=None):
        super().__init__(
            status_code=500,
            code=ErrorCodes.ERR_CONFIG_FILE,
            message=f"{message}",
            detail=detail,
        )


def handle_base_error(logger: Logger, exception: BaseError):
    logger.opt(exception=exception).exception(exception.message)
    raise exception


def handle_http_errors(logger: Logger, error_resp: ErrorResponse):
    if not error_resp:
        return
    if error_resp.error.code == ErrorCodes.ERR_INTERNAL_SERVER:
        raise_internal_server_error(logger=logger, message=error_resp.error.message)
    elif error_resp.error.code == ErrorCodes.ERR_NOT_FOUND:
        raise_not_found_error(
            logger=logger,
            entity_type=error_resp.error.detail.get("entity_type"),
            entity_key=error_resp.error.detail.get("entity_key"),
        )
    elif error_resp.error.code == ErrorCodes.ERR_REQUEST_VALIDATION:
        raise_request_validation_error(logger=logger, message=error_resp.error.message)
    elif error_resp.error.code == ErrorCodes.ERR_SCHEMA_VALIDATION:
        raise_schema_validation_error(logger=logger, message=error_resp.error.message)
    elif error_resp.error.code == ErrorCodes.ERR_DATA_INTEGRITY:
        raise_data_integrity_error(
            logger=logger,
            entity_type=error_resp.error.detail.get("entity_type"),
            entity_key=error_resp.error.detail.get("entity_key"),
        )
    elif error_resp.error.code == ErrorCodes.ERR_AUTHENTICATION_FAILED:
        raise_authentication_failed_error(
            logger=logger, message=error_resp.error.message
        )
    else:
        raise BaseError(**error_resp.error)


def handle_schema_validation_error(logger: Logger, exception: ValidationError):
    exc = SchemaValidationError(exception.errors())
    logger.opt(exception=exc).exception(
        f"Schema validation error: {exception.errors()}"
    )
    raise exc from exception


def handle_request_timeout_error(logger: Logger, exception: Exception, detail=None):
    exc = RequestTimeoutError(message="Network request timed out", detail=detail)
    logger.opt(exception=exc).exception(f"Request timeout error: {str(exception)}")
    raise exc from exception


def raise_schema_validation_error(
    logger: Logger, message: Optional[str] = None, detail=None
):

    exception = SchemaValidationError(message=message, detail=detail)
    logger.opt(depth=1).exception(
        f"Schema validation error: {message}", exception=exception
    )
    raise exception


def raise_not_found_error(
    logger: Logger,
    entity_type: str,
    entity_key: str = None,
    reference_entity_type: str = None,
    reference_entity_key: str = None,
    comment: Optional[str] = None,
):
    if entity_type and reference_entity_type and reference_entity_key:
        message = f"{entity_type} not found for {reference_entity_type} ({reference_entity_key})"
    if entity_key:
        message = f"{entity_type} ({entity_key}) not found"
    else:
        message = f"{entity_type} not found"

    if comment:
        message = f"{message} - {comment}"
    exc = NotFoundError(entity_type, entity_key)
    logger.opt(exception=exc).exception(message)
    raise exc


def raise_request_validation_error(logger: Logger, message: Optional[str] = None):
    exc = RequestValidationCustomError(message=message)
    logger.opt(exception=exc).exception(f"{message}")
    raise exc


def raise_workflow_processing_error(logger: Logger, message: Optional[str] = None):
    try:
        raise WorkflowProcessingError(message=message)
    except WorkflowProcessingError:
        logger.exception(f"Error processing workflow: {message}")
        raise


def raise_not_implemented_error(logger: Logger, message: Optional[str] = None):
    try:
        raise NotImplementError(message=message)
    except NotImplementError:
        logger.error(f"Not implemented: {message}")
        raise


def raise_internal_server_error(logger: Logger, message: Optional[str] = None):
    try:
        raise InternalServerError(message=message)
    except InternalServerError:
        logger.error(f"Internal server error: {message}")
        raise


def raise_data_integrity_error(
    logger: Logger,
    entity_type: str,
    entity_key: str,
    detail=None,
):
    exc = DataIntegrityError(
        entity_type,
        entity_key,
        detail=detail or {"entity_type": entity_type, "entity_key": entity_key},
    )
    logger.opt(exception=exc).exception(f"{entity_type} ({entity_key}) already exists")
    raise exc


def raise_authentication_failed_error(logger: Logger, message: Optional[str] = None):
    try:
        raise AuthenticationFailedError(message=message)
    except AuthenticationFailedError:
        logger.error(f"Authentication failed: {message}")
        raise


def raise_wrapper_function_error(logger: Logger, message: Optional[str] = None):
    try:
        raise WrapperFunctionError(message=message)
    except WrapperFunctionError:
        logger.error(f"Wrapper function error: {message}")
        raise


def raise_config_file_error(logger: Logger, message: Optional[str] = None):
    try:
        raise ConfigFileError(message=message)
    except ConfigFileError:
        logger.error(f"Config file error: {message}")
        raise
